# -*- coding: utf-8; -*-

# This file is part of python-ina219.
#
# python-ina219 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-ina219 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-ina219.  If not, see <http://www.gnu.org/licenses/>

# 2018-01-29 David Goncalves <david@lestat.st>
#            Python class to communicate with an INA219 chip on I2C bus.
#            It's a Texas Instruments voltage and current sensor.

import smbus
import sys


class ina219:

    # Define some contstans to address various parts of the chip

    # Internal registers
    REG_CONFIG = 0x00
    REG_SHUNT_VOLTAGE = 0x01
    REG_BUS_VOLTAGE = 0x02
    REG_POWER = 0x03
    REG_CURRENT = 0x04
    REG_CALIBRATION = 0x05

    # Configuration register parameters

    # RST = Reset the chip like a power-on and set values to their defaults
    CONF_RST = 1 << 15

    # BRNG = Bus Voltage Range
    CONF_BRNG_MASK = 0x1 << 13
    CONF_BRNG_16V = 0x0              # Bus voltage set to 16v
    CONF_BRNG_32V = 0x1 << 13        # Bus voltage set to 32v (default)

    # PGA = Programmable Gain
    CONF_PGA_MASK = 0x3 << 11
    CONF_PGA_DIV1 = 0x0              # Divide by 1 => Range -+ 40mV
    CONF_PGA_DIV2 = 0x1 << 11        # Divide by 2 => Range -+ 80mv
    CONF_PGA_DIV4 = 0x2 << 11        # Divide by 4 => Range -+ 160mV
    CONF_PGA_DIV8 = 0x3 << 11        # Divide by 8 => Range -+ 320mV (default)

    # BADC = Bus ADC resolution and number of samples for averaging measurements
    CONF_BADC_MASK = 0xf << 7
    CONF_BADC_9BIT = 0x0             # Bus ADC resolution = 9bit
    CONF_BADC_10BIT = 0x1 << 7       # Bus ADC resolution = 10bit
    CONF_BADC_11BIT = 0x2 << 7       # Bus ADC resolution = 11bit
    CONF_BADC_12BIT = 0x3 << 7       # Bus ADC resolution = 12bit (default)
    CONF_BADC_2SAMPLES = 0x9 << 7    # Use 2 samples for averaging
    CONF_BADC_4SAMPLES = 0xa << 7    # Use 4 samples for averaging
    CONF_BADC_8SAMPLES = 0xb << 7    # Use 8 samples for averaging
    CONF_BADC_16SAMPLES = 0xc << 7   # Use 16 samples for averaging
    CONF_BADC_32SAMPLES = 0xd << 7   # Use 32 samples for averaging
    CONF_BADC_64SAMPLES = 0xe << 7   # Use 64 samples for averaging
    CONF_BADC_128SAMPLES = 0xf << 7  # Use 128 samples for averaging

    # SADC = Shunt ADC resolution and number of samples for averaging measurements
    CONF_SADC_MASK = 0xf << 3
    CONF_SADC_9BIT = 0x0             # Shunt ADC resolution = 9bit
    CONF_SADC_10BIT = 0x1 << 3       # Shunt ADC resolution = 10bit
    CONF_SADC_11BIT = 0x2 << 3       # Shunt ADC resolution = 11bit
    CONF_SADC_12BIT = 0x3 << 3       # Shunt ADC resolution = 12bit (default)
    CONF_SADC_2SAMPLES = 0x9 << 3    # Use 2 samples for averaging
    CONF_SADC_4SAMPLES = 0xa << 3    # Use 4 samples for averaging
    CONF_SADC_8SAMPLES = 0xb << 3    # Use 8 samples for averaging
    CONF_SADC_16SAMPLES = 0xc << 3   # Use 16 samples for averaging
    CONF_SADC_32SAMPLES = 0xd << 3   # Use 32 samples for averaging
    CONF_SADC_64SAMPLES = 0xe << 3   # Use 64 samples for averaging
    CONF_SADC_128SAMPLES = 0xf << 3  # Use 128 samples for averaging

    # MODE = Operating mode
    CONF_MODE_MASK = 0x7
    CONF_MODE_POWER_DOWN = 0x0       # Power down chip (go to low power mode)
    CONF_MODE_SHUNT_TRIG = 0x1       # Shunt voltage, triggered
    CONF_MODE_BUS_TRIG = 0x2         # Bus voltage, triggered
    CONF_MODE_SHUNT_BUS_TRIG = 0x3   # Shunt and Bus voltage, triggered
    CONF_MODE_ADC_OFF = 0x4          # Disable ADC converter
    CONF_MODE_SHUNT_CONT = 0x5       # Shunt voltage, continuous
    CONF_MODE_BUS_CONT = 0x6         # Bus voltage, confinuous
    CONF_MODE_SHUNT_BUS_CONT = 0x7   # Shunt and Bus voltage, continuous (default)

    chip_config = 0
    current_lsb = 0

    def __init__(self, address=0x40, bus=1, debug=False):
        self._debug = debug
        self.i2c_addr = address
        self.i2c_bus = smbus.SMBus(bus)

        self._debug_message("Accessing INA219 at address 0x%x on bus %s" % (address, bus))
        self._read_config()

    def _debug_message(self, message="", error=False):
        """
        Prints debug messages to console when debuging mode is activated
        """

        if (self._debug is True) or (error is True):
            print("%s %s" % (("[DEBUG]" if error is False else "[ERROR]"), message))
            sys.stdout.flush()
            if (error is True):
                raise Exception(message)

    def _write_data(self, address, data):
        """
        Private method to write data to the INA219
        """

        self._debug_message("> Writing %s bytes to register 0x%x [%s]" % (len(data), address, ' '.join(["0x%02X" % x for x in data])))
        self.i2c_bus.write_i2c_block_data(self.i2c_addr, address, data)

    def _read_data(self, address, size):
        """
        Private method to read data from the INA219
        """

        self._debug_message("> Reading %s bytes from register 0x%x" % (size, address))
        data = self.i2c_bus.read_i2c_block_data(self.i2c_addr, address, size)
        self._debug_message("> INA219 response : [%s]" % ' '.join(["0x%02X" % x for x in data]))
        return(data)

    def _read_config(self):
        """
        Read and update the local var containing the INA219 configuration
        """

        self._debug_message("Reading configuration from INA219")
        data = self._read_data(self.REG_CONFIG, 2)
        self.chip_config = (data[0] << 8) + data[1]

    def _two_comp(self, value, bitWidth):
        """
        Return the two complement's value of a value using 'bitWidth' resolution.

        This function comes from the following post in StackOverflow
        https://stackoverflow.com/questions/1604464/twos-complement-in-python
        """

        if value >= 2**bitWidth:
            # This catches when someone tries to give a value that is out of range
            raise ValueError("Value: {} out of range of {}-bit value.".format(value, bitWidth))
        else:
            return value - int((value << 1) & 2**bitWidth)

    def SoftReset(self):
        """
        Perform a soft reset of the INA219 by writing '1' to the CONF_RST bit of the config register
        """

        self._debug_message("Performing a soft reset of the INA219")
        config = self.CONF_RST
        self._write_data(self.REG_CONFIG, [config >> 8, config & 0xFF])
        # Reset the current_lsb variable as we need new calibration after reset
        self.current_lsb = 0

    def GetBusVoltage(self):
        """
        Read Bus Voltage register
        This method will return a dict containing 3 keys :
          - value (int)     = The Bus Voltage in millivolts
          - overflow (bool) = Flag (OVF) reported by the chip to indicate that the voltage is out of range
          - ready (bool)    = Flag (CNVR) indicating that the chip as finished its conversion
        """

        self._debug_message("Reading Bus Voltage register")
        data = self._read_data(self.REG_BUS_VOLTAGE, 2)
        raw = (data[0] << 8) + data[1]

        result = {}
        result["overflow"] = bool(raw & 1)       # OVF is on bit 0
        result["ready"] = bool((raw >> 1) & 1)   # CNVR is on bit 1
        result["value"] = (raw >> 3) * 4         # Voltage is from bit 3 to 15 (with LSB = 4mV)
        self._debug_message("Measured Bus voltage is %smV (OVF=%s CNVR=%s)" % (result["value"], int(result["overflow"]), int(result["ready"])))
        return(result)

    def GetShuntVoltage(self):
        """
        Read shunt voltage register and return value as millivolts (according to PGA config)
        """

        self._debug_message("Reading Shunt Voltage register")
        data = self._read_data(self.REG_SHUNT_VOLTAGE, 2)
        raw = (data[0] << 8) + data[1]

        # Extract bit depth from config register (PG0,PG1)
        depth = 13 + (self.chip_config >> 11) & ~(1 << 2)

        raw = (self._two_comp(raw, depth) * 0.00001) * 100     # Return value in millivolts (MSB is 10µV)
        self._debug_message("Measured Shunt voltage is %smV (resolution %s bits)" % (raw, depth))
        return(raw)

    def GetCurrent(self):
        """
        Read the current register and return value in milliamperes according to the calibration setup
        """

        self._debug_message("Reading Current register")
        data = self._read_data(self.REG_CURRENT, 2)
        raw = self._two_comp((data[0] << 8) + data[1], 16)
        self._debug_message("Measured current is %smA" % (raw * self.current_lsb * 1000))
        return(raw * self.current_lsb * 1000)

    def GetPower(self):
        """
        Read the power register and return value in milliwatts according to the calibration setup
        """

        self._debug_message("Reading Power register")
        data = self._read_data(self.REG_POWER, 2)
        raw = self._two_comp((data[0] << 8) + data[1], 16)
        self._debug_message("Measured power is %smW" % (raw * 20 * self.current_lsb * 1000))
        return(raw * 20 * self.current_lsb * 1000)

    def Calibrate(self, r_shunt=0, i_max=0):
        """
        Calibrate the INA219 (using the calibration register)
        Arguments are :
          - r_shunt (int) = Shunt resistor value (in Ohms)
          - i_max (float) = Expected max current (in Amperes)
        """

        # TODO: Check boundaries to avoid calibration > 32767

        self._debug_message("Calibrating INA219 with Imax=%sA RShunt=%sΩ" % (i_max, r_shunt))
        self.current_lsb = i_max / 2**15
        calibration = int(0.04096 / (self.current_lsb * r_shunt)) << 1

        self._write_data(self.REG_CALIBRATION, [calibration >> 8, calibration & 0xFF])

    def TriggerMeasurement(self):
        """
        This method triggers a new measurement by writing on the Calibration register.
        """

        self._debug_message("Triggering a new measurement")
        self._read_config()

        # Only write to register if the chip is configured using a triggered mode
        if (self.chip_config & self.CONF_MODE_MASK) == self.CONF_MODE_SHUNT_TRIG or \
           (self.chip_config & self.CONF_MODE_MASK) == self.CONF_MODE_BUS_TRIG or \
           (self.chip_config & self.CONF_MODE_MASK) == self.CONF_MODE_SHUNT_BUS_TRIG:

            data = self._read_data(self.REG_CALIBRATION, 2)
            self._write_data(self.REG_CALIBRATION, data)

        else:
            self._debug_message("INA219 not configured using a triggered mode, skipping...")

    def Configure(self, config=None, single_value=None, mask=0):
        """
        Configure the INA219 by writting on the CONFIG register.
        This method can act using two different methods.

        - set the 'config' paremeter containing all parameters
        - update a single 'value' on existing config (using mask)
        """

        if ((config is None) and (single_value is None)) or \
           ((config is not None) and (single_value is not None)):
            raise ValueError("Please choose 'config' or 'single_value' mode for ConfigureChip")

        if config is not None:
            self._debug_message("Configuring INA219 using full parameters")
            if config > 0xFFFF:
                raise ValueError("config parameter shouldn't be more that 2 bytes (%s)" % config)

            self._write_data(self.REG_CONFIG, [config >> 8, config & 0xFF])

        if single_value is not None:
            self._debug_message("Updating single value on INA219 configuration")
            if single_value > 0xFFFF:
                raise ValueError("config parameter shouldn't be more that 2 bytes (%s)" % single_value)

            # Refresh the current chip config before updating it
            self._read_config()

            data = (self.chip_config & (~ mask)) | single_value
            self._write_data(self.REG_CONFIG, [data >> 8, data & 0xFF])
            self.chip_config = data
