# 2019-03-05 David Goncalves <david@lestat.st>
#            Python class to communicate with an INA219 chip on I2C bus.
#            It's a Texas Instruments voltage and current sensor.
#            > Code ported to micropython

import sys
import time
import ustruct
from micropython import const


class uina219:

    REG_CONFIG = const(0x00)
    REG_SHUNT_VOLTAGE = const(0x01)
    REG_BUS_VOLTAGE = const(0x02)
    REG_POWER = const(0x03)
    REG_CURRENT = const(0x04)
    REG_CALIBRATION = const(0x05)

    CONF_RST = const(1 << 15)

    CONF_BRNG_MASK = const(0x1 << 13)
    CONF_BRNG_16V = const(0x0)
    CONF_BRNG_32V = const(0x1 << 13)

    CONF_PGA_MASK = const(0x3 << 11)
    CONF_PGA_DIV1 = const(0x0)
    CONF_PGA_DIV2 = const(0x1 << 11)
    CONF_PGA_DIV4 = const(0x2 << 11)
    CONF_PGA_DIV8 = const(0x3 << 11)

    CONF_BADC_MASK = const(0xf << 7)
    CONF_BADC_9BIT = const(0x0)
    CONF_BADC_10BIT = const(0x1 << 7)
    CONF_BADC_11BIT = const(0x2 << 7)
    CONF_BADC_12BIT = const(0x3 << 7)
    CONF_BADC_2SAMPLES = const(0x9 << 7)
    CONF_BADC_4SAMPLES = const(0xa << 7)
    CONF_BADC_8SAMPLES = const(0xb << 7)
    CONF_BADC_16SAMPLES = const(0xc << 7)
    CONF_BADC_32SAMPLES = const(0xd << 7)
    CONF_BADC_64SAMPLES = const(0xe << 7)
    CONF_BADC_128SAMPLES = const(0xf << 7)

    CONF_SADC_MASK = const(0xf << 3)
    CONF_SADC_9BIT = const(0x0)
    CONF_SADC_10BIT = const(0x1 << 3)
    CONF_SADC_11BIT = const(0x2 << 3)
    CONF_SADC_12BIT = const(0x3 << 3)
    CONF_SADC_2SAMPLES = const(0x9 << 3)
    CONF_SADC_4SAMPLES = const(0xa << 3)
    CONF_SADC_8SAMPLES = const(0xb << 3)
    CONF_SADC_16SAMPLES = const(0xc << 3)
    CONF_SADC_32SAMPLES = const(0xd << 3)
    CONF_SADC_64SAMPLES = const(0xe << 3)
    CONF_SADC_128SAMPLES = const(0xf << 3)

    CONF_MODE_MASK = const(0x7)
    CONF_MODE_POWER_DOWN = const(0x0)
    CONF_MODE_SHUNT_TRIG = const(0x1)
    CONF_MODE_BUS_TRIG = const(0x2)
    CONF_MODE_SHUNT_BUS_TRIG = const(0x3)
    CONF_MODE_ADC_OFF = const(0x4)
    CONF_MODE_SHUNT_CONT = const(0x5)
    CONF_MODE_BUS_CONT = const(0x6)
    CONF_MODE_SHUNT_BUS_CONT = const(0x7)

    chip_config = 0
    current_lsb = 0

    def __init__(self, bus, address=0x40):
        self.i2c_bus = bus
        self.i2c_addr = address
        self._read_config()

    def _write_data(self, address, data):
        self.i2c_bus.writeto_mem(self.i2c_addr, address, bytearray(data))

    def _read_data(self, address, size):
        return(self.i2c_bus.readfrom_mem(self.i2c_addr, address, size))

    def _two_comp(self, value, bitWidth):
        if value >= 2**bitWidth:
            # This catches when someone tries to give a value that is out of range
            raise ValueError("Value: {} out of range of {}-bit value.".format(value, bitWidth))
        else:
            return(value - int((value << 1) & 2**bitWidth))

    def _read_config(self):
        self.chip_config = ustruct.unpack('>H', self._read_data(self.REG_CONFIG, 2))[0]

    def SoftReset(self):
        self._write_data(self.REG_CONFIG, ustruct.pack(">H", self.CONF_RST))
        self.current_lsb = 0

    def GetBusVoltage(self):
        raw = ustruct.unpack('>H', self._read_data(self.REG_BUS_VOLTAGE, 2))[0]
        return({"overflow": bool(raw & 1), "ready": bool((raw >> 1) & 1), "value": (raw >> 3) * 4})

    def GetShuntVoltage(self):
        raw = ustruct.unpack('>H', self._read_data(self.REG_SHUNT_VOLTAGE, 2))[0]
        depth = 13 + (self.chip_config >> 11) & ~(1 << 2)
        return((self._two_comp(raw, depth) * 0.00001) * 100)

    def GetCurrent(self):
        raw = self._two_comp(ustruct.unpack('>H', self._read_data(self.REG_CURRENT, 2))[0], 16)
        return(raw * self.current_lsb * 1000)

    def GetPower(self):
        raw = self._two_comp(ustruct.unpack('>H', self._read_data(self.REG_POWER, 2))[0], 16)
        return(raw * 20 * self.current_lsb * 1000)

    def Calibrate(self, r_shunt=0, i_max=0):
        self.current_lsb = i_max / 2**15
        self._write_data(self.REG_CALIBRATION, ustruct.pack('>H', int(0.04096 / (self.current_lsb * r_shunt)) << 1))

    def TriggerMeasurement(self):
        self._read_config()
        if (self.chip_config & self.CONF_MODE_MASK) == self.CONF_MODE_SHUNT_TRIG or \
           (self.chip_config & self.CONF_MODE_MASK) == self.CONF_MODE_BUS_TRIG or \
           (self.chip_config & self.CONF_MODE_MASK) == self.CONF_MODE_SHUNT_BUS_TRIG:
            data = self._read_data(self.REG_CALIBRATION, 2)
            self._write_data(self.REG_CALIBRATION, data)
            return(True)

        else:
            return(False)

    def Configure(self, config=None, single_value=None, mask=0):
        if ((config is None) and (single_value is None)) or \
           ((config is not None) and (single_value is not None)):
            raise ValueError("Please choose 'config' or 'single_value' mode for ConfigureChip")

        if config is not None:
            if config > 0xFFFF:
                raise ValueError("config parameter shouldn't be more that 2 bytes (%s)" % config)

            self._write_data(self.REG_CONFIG, ustruct.pack('>H', config))

        if single_value is not None:
            if single_value > 0xFFFF:
                raise ValueError("config parameter shouldn't be more that 2 bytes (%s)" % single_value)

            self._read_config()
            self._write_data(self.REG_CONFIG, ustruct.pack('>H', ((self.chip_config & (~ mask)) | single_value)))
            self._read_config()
